#include "model/CItem.h"

#include "config/Config.h"
#include "db/DBConn.h"
#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "memory/Allocator.h"

// L2CacheD 0x0049D140
long CItem::s_nAlloc = ::InterlockedIncrement(&MemoryObject::s_refCountSlots);  // L2CacheD 0x00AFE32C

CItemeMap g_itemCache;  // L2CacheD 0x00AFE2DC

// L2CacheD 0x00425120
CItem::CItem(uint32_t itemId)
    : MemoryObject()
    , m_transactMode(TransactType_Invalid)
    , m_itemId(itemId)
    , m_ownerId(0)
    , m_type(0)
    , m_amount(0)
    , m_enchant(0)
    , m_damaged(0)
    , m_bless(0)
    , m_ident(0)
    , m_wished(0)
    , m_warehouse(WareHouseType_Inventory)
    , field_60(true)
    , m_inTransactions(0)
    , m_deleted(0)
    , m_auditRelated(-1)
    , field_76(0)
    , field_80(0)
{
    guard(L"CItem::CItem(uint32_t nItemID)", L"CItem::CItem(uint32_t nItemID) (exit)", TEXT(__FILE__), __LINE__);

    unguard();
}

// L2CacheD 0x004253E0
CItem::CItem(uint32_t itemId, uint32_t ownerId, int type, uint32_t amount, int enchant, int damaged, int bless, int ident, int wished, WareHouseType warehouse)
    : MemoryObject()
    , m_transactMode(TransactType_Invalid)
    , m_itemId(itemId)
    , m_ownerId(0)
    , m_type(0)
    , m_amount(0)
    , m_enchant(0)
    , m_damaged(0)
    , m_bless(0)
    , m_ident(0)
    , m_wished(0)
    , m_warehouse(WareHouseType_Inventory)
    , field_60(true)
    , m_inTransactions(0)
    , m_deleted(0)
    , m_auditRelated(-1)
    , field_76(0)
    , field_80(0)
{
    guard(L"CItem::CItem(...)", L"CItem::CItem(...) (exit)", TEXT(__FILE__), __LINE__);
    Set(ownerId, type, amount, enchant, damaged, bless, ident, wished, warehouse);
    unguard();
}

// L2CacheD inlined inside 0x004251D0
CItem::~CItem()
{
}

// L2CacheD 0x004247D0
void* CItem::operator new(size_t size)
{
    void* memory = g_allocator.allocate(size, 1, L"CItem");

    ::InterlockedIncrement(&Allocator::g_allocatedObjects[CItem::s_nAlloc]);

    Allocator::g_allocatedObjName[CItem::s_nAlloc] = "CItem";
    Allocator::g_allocatedMemory[CItem::s_nAlloc] = sizeof(CItem);

    return memory;
}

// L2CacheD inlined in 0x004251D0
void CItem::operator delete(void* /*ptr*/)
{
    g_winlog.Add(LOG_ERROR, L"Never Call Oprator Delete!!! CItem at file[%s], line[%d]", TEXT(__FILE__), __LINE__);
    CallStack::ReportCallStack();
}

// L2CacheD 0x00424840
void CItem::Init()
{
    guard(L"void CItem::Init()", L"void CItem::Init() (exit)", TEXT(__FILE__), __LINE__);

    if (Config::s_itemClearOnStart)
    {
        g_winlog.Add(LOG_INF, L"Clearing Unowned items...");
        DBConn sql;
        sql.Execute(L"EXEC lin_DeleteNotOwnedItem");
        Config::s_itemClearOnStart = false;
        g_winlog.Add(LOG_INF, L"Unowned items are cleared.");
    }

    unguard();
}

// L2CacheD 0x00425DF0
CItemSP CItem::Load(uint32_t itemId)
{
    guard(L"CItem* CItem::Load(uint32_t nItemID)", L"CItem* CItem::Load(uint32_t nItemID) (exit)", TEXT(__FILE__), __LINE__);

    CItemSP result = g_itemCache.FindObject(itemId);
    if (result != NULL)
    {
        unguard();
        return result;
    }

    uint32_t ownerId = 0;
    int type = 0;
    uint32_t amount = 0;
    int enchant = 0;
    int damaged = 0;
    int bless = 0;
    int ident = 0;
    int wished = 0;
    int warehouse = 0;

    DBConn sql;
    sql.Bind(&ownerId);
    sql.Bind(&type);
    sql.Bind(&amount);
    sql.Bind(&enchant);
    sql.Bind(&damaged);
    sql.Bind(&bless);
    sql.Bind(&ident);
    sql.Bind(&wished);
    sql.Bind(&warehouse);
    if (sql.Execute(L"EXEC lin_LoadItemByItemId %d", itemId) && sql.Fetch())
    {
        CItem* loadedItem = new CItem(itemId, ownerId, type, amount, enchant, damaged, bless, ident, wished, static_cast<WareHouseType>(warehouse));
        g_itemCache.Insert(itemId, loadedItem, NULL, -1);

        result.SetObject(loadedItem, ORT_CS_POINTER);
        loadedItem->Release(__FILE__, __LINE__, ORT_CREATE_DELETE, 0);
    }

    unguard();
    return result;
}

// L2CacheD 0x00424FF0
void CItem::Release(const char* file, int line, ORTs /*nWorkType*/, bool /*flag*/)
{
    long refCountAfterRelease = ::InterlockedDecrement(&m_nRefCount);
    if (refCountAfterRelease > 0)
    {
        return;
    }

    if (refCountAfterRelease != 0)
    {
        if (refCountAfterRelease > -1)
        {
            return;
        }

        g_winlog.Add_Ansi(LOG_ERROR, "Invalid Memory (%d) on CItem(%d)", refCountAfterRelease, m_identity);
        if (file)
        {
            g_winlog.Add_Ansi(LOG_ERROR, "           Caller : file[%s], line[%d]", file, line);
        }

        ReportOrt(0, 0);
        CallStack::ReportCallStack();

        return;
    }

    if (inTimer())
    {
        if (file)
        {
            g_winlog.Add_Ansi(LOG_ERROR, "deleting in-timer-object![CItem] at file[%s], line[%d]", file, line);
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CItem::s_nAlloc]);
        }
        else
        {
            g_winlog.Add_Ansi(LOG_ERROR, "deleting in-timer-object![CItem]");
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CItem::s_nAlloc]);
        }
        return;
    }

    this->~CItem();

    g_allocator.deallocate(this);

    ::InterlockedDecrement(&Allocator::g_allocatedObjects[CItem::s_nAlloc]);
}

// L2CacheD 0x00425210
CItemSP CItem::Set(uint32_t ownerId, int type, uint32_t amount, int enchant, int damaged, int bless, int ident, int wished, WareHouseType warehouse)
{
    guard(L"CItem* CItem::Set(...)", L"CItem* CItem::Set(...) (exit)", TEXT(__FILE__), __LINE__);
    m_ownerId = ownerId;
    m_type = type;
    m_amount = amount;
    m_enchant = enchant;
    m_damaged = damaged;
    m_bless = bless;
    m_ident = ident;
    m_wished = wished;
    m_warehouse = warehouse;

    CItemSP result(this, ORT_CS_POINTER);
    unguard();

    return result;
}

// L2CacheD 0x004254E0
CItemSP CItem::Copy() const
{
    guard(L"CItem::Copy()", L"CItem::Copy() (exit)", TEXT(__FILE__), __LINE__);

    CItem* copy = new CItem(m_itemId, m_ownerId, m_type, m_amount, m_enchant, m_damaged, m_bless, m_ident, m_wished, m_warehouse);
    CItemSP result(copy, ORT_CS_POINTER);
    copy->Release(__FILE__, __LINE__, ORT_CREATE_DELETE, 0);

    unguard();
    return result;
}

// L2CacheD 0x00425BE0
bool CItem::Erase()
{
    guard(L"bool CItem::Erase()", L"bool CItem::Erase() (exit)", TEXT(__FILE__), __LINE__);
    g_itemCache.DeleteObject(m_itemId, NULL, -1);
    unguard();

    return true;
}

// L2CacheD 0x00425C50
bool CItem::Delete(bool hardDelete)
{
    guard(L"bool CItem::Delete()", L"bool CItem::Delete() (exit)", TEXT(__FILE__), __LINE__);
    if (!hardDelete)
    {
        g_itemCache.DeleteObject(m_itemId, NULL, -1);
        unguard();
        return true;
    }

    int rowCount = 0;
    DBConn sql;
    sql.Bind(&rowCount);
    if (!sql.Execute(L"EXEC lin_DeleteItem %d", m_itemId))
    {
        g_winlog.Add(LOG_ERROR, L"Item delete Error (execute), id[%d]", m_itemId);
        unguard();
        return false;
    }

    if (!sql.Fetch())
    {
        g_winlog.Add(LOG_ERROR, L"Item delete Error (fetch), id[%d]", m_itemId);
        unguard();
        return false;
    }

    g_winlog.Add(LOG_REQUEST, L"%s, %d", L"EXEC lin_DeleteItem %d", m_itemId);

    g_itemCache.DeleteObject(m_itemId, NULL, -1);
    unguard();
    return true;
}

// L2CacheD 0x00426070
void CItem::CacheItem()
{
    guard(L"void CItem::CacheItem()", L"void CItem::CacheItem() (exit)", TEXT(__FILE__), __LINE__);

    g_itemCache.Insert(m_itemId, this, NULL, -1);

    unguard();
}

// L2CacheD 0x004260F0
bool CItem::CancelTransation(CItem* rollbackItem)
{
    guard(L"CItem::CancelTransation()", L"CItem::CancelTransation() (exit)", TEXT(__FILE__), __LINE__);

    if (m_inTransactions == 0)
    {
        g_winlog.Add(LOG_ERROR, L"Rollback Item Error, id[%d], user[%d]. Not in transation.", m_itemId, m_ownerId);
        unguard();
        return false;
    }

    TransactType transactMode = rollbackItem->GetTrasactMode();
    if (transactMode == TransactType_Invalid)
    {
        g_winlog.Add(LOG_ERROR, L"Item transaction ERROR [%s][%d]", TEXT(__FILE__), __LINE__);
        unguard();
        return false;
    }

    if (transactMode == TransactType_Rollback)
    {
        SetDBID(rollbackItem->ItemID());
        Set(rollbackItem->OwnerID(), rollbackItem->ItemType(), rollbackItem->Amount(), rollbackItem->Enchant(), rollbackItem->Damaged(), rollbackItem->Bless(), rollbackItem->Ident(), rollbackItem->Wished(), rollbackItem->Warehouse());
        m_deleted = false;
        unguard();
        return true;
    }

    if (transactMode != TransactType_RollbackDelete)
    {
        g_winlog.Add(LOG_ERROR, L"[%s][%d] Rollback ERROR", TEXT(__FILE__), __LINE__);
        unguard();
        return false;
    }

    if (!rollbackItem->Delete(true))
    {
        g_winlog.Add(LOG_ERROR, L"Rollback item delete Error, id[%d], user[%d]", m_itemId, m_ownerId);
    }

    Delete(true);

    unguard();
    return true;
}

// L2CacheD 0x004262F0
bool CItem::CommitTransation(CItem* other)
{
    throw "Not implemented!";  // TODO: port me!
}

// L2CacheD 0x00424940
bool CItem::TransactDelete()
{
    guard(L"bool CItem::TransactDelete()", L"bool CItem::TransactDelete() (exit)", TEXT(__FILE__), __LINE__);

    bool result = false;
    if (m_inTransactions == 0)
    {
        result = false;
    }
    else
    {
        m_deleted = true;
        result = true;
    }

    unguard();
    return result;
}

// L2CacheD 0x00424A00
bool CItem::Save()
{
    guard(L"bool CItem::Save() const", L"bool CItem::Save() const (exit)", TEXT(__FILE__), __LINE__);

    DBConn sql;
    bool result = sql.Execute(L"EXEC lin_UpdateUserItem %d, %d, %d, %d, %d, %d, %d, %d, %d, %d", m_ownerId, m_type, m_amount, m_enchant, m_damaged, m_bless, m_ident, m_wished, m_warehouse, m_itemId);
    unguard();

    return result;
}

// L2CacheD 0x00425310
CItemSP CItem::SetTransation()
{
    guard(L"CItem::SetTransation()", L"CItem::SetTransation() (exit)", TEXT(__FILE__), __LINE__);

    ++m_inTransactions;
    m_deleted = false;

    CItemSP result(this, ORT_CS_POINTER);
    unguard();

    return result;
}

// L2CacheD 0x00424B00
uint32_t CItem::ItemID() const
{
    guard(L"uint32_t CItem::ItemID() const", L"uint32_t CItem::ItemID() const (exit)", TEXT(__FILE__), __LINE__);
    uint32_t id = m_itemId;
    unguard();
    return id;
}

// L2CacheD 0x004249C0
void CItem::SetDBID(uint32_t itemId)
{
    guard(L"CItem* CItem::SetDBID(uint32_t nItemID)", L"CItem* CItem::SetDBID(uint32_t nItemID) (exit)", TEXT(__FILE__), __LINE__);
    m_itemId = itemId;
    unguard();
}

// L2CacheD 0x00424D40
WareHouseType CItem::Warehouse() const
{
    guard(L"int CItem::WareHouse() const", L"int CItem::WareHouse() const (exit)", TEXT(__FILE__), __LINE__);
    WareHouseType warehouse = m_warehouse;
    unguard();
    return warehouse;
}

// L2CacheD 0x00424D80
void CItem::SetWarehouse(WareHouseType warehouse)
{
    guard(L"CItem::SetWarehouse(...)", L"CItem::SetWarehouse(...) (exit)", TEXT(__FILE__), __LINE__);
    m_warehouse = warehouse;
    unguard();
}

// L2CacheD 0x00424DC0
TransactType CItem::GetTrasactMode() const
{
    guard(L"CItem::GetTrasactMode()", L"CItem::GetTrasactMode() (exit)", TEXT(__FILE__), __LINE__);
    TransactType mode = m_transactMode;
    unguard();
    return mode;
}

// L2CacheD 0x00424E00
void CItem::SetTrasactMode(TransactType modeTransactType)
{
    guard(L"CItem::SetTrasactMode(TransactType modeTransactType)", L"CItem::SetTrasactMode(TransactType modeTransactType) (exit)", TEXT(__FILE__), __LINE__);
    m_transactMode = modeTransactType;
    unguard();
}

// L2CacheD 0x00424BC0
uint32_t CItem::Amount() const
{
    guard(L"int CItem::Amount() const", L"int CItem::Amount() const (exit)", TEXT(__FILE__), __LINE__);
    uint32_t amount = m_amount;
    unguard();
    return amount;
}

// L2CacheD 0x00424E40
void CItem::SetAmount(uint32_t amount)
{
    guard(L"CItem::SetAmount(int nAmount)", L"CItem::SetAmount(int nAmount) (exit)", TEXT(__FILE__), __LINE__);
    if (amount < 0)
    {
        g_winlog.Add(LOG_ERROR, L"Set amount is try below 0.[][]");
        amount = 1;
    }
    m_amount = amount;
    unguard();
}

// L2CacheD 0x00424B40
uint32_t CItem::OwnerID() const
{
    guard(L"uint32_t CItem::OwnerID() const", L"uint32_t CItem::OwnerID() const (exit)", TEXT(__FILE__), __LINE__);
    uint32_t ownerId = m_ownerId;
    unguard();
    return ownerId;
}

void CItem::setOwnerID(uint32_t ownerId)
{
    m_ownerId = ownerId;
}

// L2CacheD 0x00424B80
int CItem::ItemType() const
{
    guard(L"uint32_t CItem::ItemType() const", L"uint32_t CItem::ItemType() const (exit)", TEXT(__FILE__), __LINE__);
    int type = m_type;
    unguard();
    return type;
}

// L2CacheD 0x00424C00
int CItem::Enchant() const
{
    guard(L"int CItem::Enchant() const", L"int CItem::Enchant() const (exit)", TEXT(__FILE__), __LINE__);
    int enchant = m_enchant;
    unguard();
    return enchant;
}

// L2CacheD 0x00424C40
int CItem::Damaged() const
{
    guard(L"int CItem::Damaged() const", L"int CItem::Damaged() const (exit)", TEXT(__FILE__), __LINE__);
    int damaged = m_damaged;
    unguard();
    return damaged;
}

// L2CacheD 0x00424C80
int CItem::Bless() const
{
    guard(L"int CItem::Bless() const", L"int CItem::Bless() const (exit)", TEXT(__FILE__), __LINE__);
    int bless = m_bless;
    unguard();
    return bless;
}

// L2CacheD 0x00424CC0
int CItem::Ident() const
{
    guard(L"int CItem::Ident() const", L"int CItem::Ident() const (exit)", TEXT(__FILE__), __LINE__);
    int ident = m_ident;
    unguard();
    return ident;
}

// L2CacheD 0x00424D00
int CItem::Wished() const
{
    guard(L"int CItem::Wished() const", L"int CItem::Wished() const (exit)", TEXT(__FILE__), __LINE__);
    int wished = m_wished;
    unguard();
    return wished;
}
