#include <WinSock2.h>

#include "network/CIOServer.h"

#include "network/CIOSocket.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "ui/CLog.h"
#include "threads/IOCP.h"
#include "memory/Allocator.h"

// L2CacheD 0x0049D5B0
long CIOServer::s_nAlloc = ::InterlockedIncrement(&MemoryObject::s_refCountSlots);  // L2CacheD 0x026846A4

// L2CacheD 0x00480510
CIOServer::CIOServer()
    : CIOObject()
    , m_hSocket(INVALID_SOCKET)
    , m_hAcceptEvent(WSA_INVALID_EVENT)
{
    guard(L"CIOServer::CIOServer()", L"CIOServer::CIOServer() (exit)", TEXT(__FILE__), __LINE__);
    unguard();
}

// L2CacheD 0x00480590
CIOServer::~CIOServer()
{
    guard(L"CIOServer::~CIOServer()", L"CIOServer::~CIOServer() (exit)", TEXT(__FILE__), __LINE__);

    Close();

    unguard();
}

// inlined inside L2CacheD 0x00481260
void CIOServer::operator delete(void* /*ptr*/)
{
    g_winlog.Add(LOG_ERROR, L"Never Call Oprator Delete!!! CIOServer at file[%s], line[%d]", TEXT(__FILE__), __LINE__);
    CallStack::ReportCallStack();
}

// L2CacheD 0x004803E0
void CIOServer::Release(const char* file, int line, ORTs /*nWorkType*/, bool /*flag*/)
{
    int refCountAfterRelease = ::InterlockedDecrement(&m_nRefCount);
    if (refCountAfterRelease > 0)
    {
        return;
    }

    if (refCountAfterRelease != 0)
    {
        if (refCountAfterRelease > -1)
        {
            return;
        }

        g_winlog.Add_Ansi(LOG_ERROR, "Invalid Memory (%d) on CIOServer(%d)", refCountAfterRelease, m_identity);
        if (file)
        {
            g_winlog.Add_Ansi(LOG_ERROR, "           Caller : file[%s], line[%d]", file, line);
        }

        ReportOrt(0, 0);
        CallStack::ReportCallStack();
        return;
    }

    if (inTimer())
    {
        if (file)
        {
            g_winlog.Add_Ansi(LOG_ERROR, "deleting in-timer-object![CIOServer] at file[%s], line[%d]", file, line);
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CIOServer::s_nAlloc]);
        }
        else
        {
            g_winlog.Add_Ansi(LOG_ERROR, "deleting in-timer-object![CIOServer]");
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CIOServer::s_nAlloc]);
        }
        return;
    }

    this->~CIOServer();

    g_allocator.deallocate(this);

    ::InterlockedDecrement(&Allocator::g_allocatedObjects[CIOServer::s_nAlloc]);
}

// L2CacheD 0x0047F8D0
void CIOServer::OnWaitCallback()
{
    guard(L"void CIOServer::OnWaitCallback()", L"void CIOServer::OnWaitCallback() (exit)", TEXT(__FILE__), __LINE__);

    ::WSAResetEvent(m_hAcceptEvent);
    ::PostQueuedCompletionStatus(IOCP::g_hCompletionPort, 0, (ULONG_PTR)this, NULL);

    unguard();
}

// L2CacheD 0x00480610
void CIOServer::OnIOCallback(BOOL /*bSuccess*/, DWORD /*dwTransferred*/, LPOVERLAPPED /*lpOverlapped*/)
{
    guard(L"void CIOServer::OnIOCallback(BOOL bSuccess, DWORD dwTransferred, LPOVERLAPPED lpOverlapped)", L"void CIOServer::OnIOCallback(BOOL bSuccess, DWORD dwTransferred, LPOVERLAPPED lpOverlapped) (exit)", TEXT(__FILE__), __LINE__);

    sockaddr_in clientAddress;
    int clientAddressLength = sizeof(clientAddress);
    SOCKET newSocket = ::accept(m_hSocket, (struct sockaddr*)&clientAddress, &clientAddressLength);

    if (newSocket == INVALID_SOCKET)
    {
        if (WSAGetLastError() == WSAEWOULDBLOCK)
        {
            unguard();
            return;
        }

        if (m_hSocket != INVALID_SOCKET)
        {
            int err = ::WSAGetLastError();
            g_winlog.Add_Ansi(LOG_ERROR, "accept error: %d", err);
        }

        unguard();
        return;
    }

    CIOSocket* pSocket = CreateSocket(newSocket, &clientAddress);
    if (pSocket != NULL)
    {
        pSocket->Initialize();
    }
    else
    {
        ::closesocket(newSocket);
        g_winlog.Add_Ansi(LOG_ERROR, "Socket Create fail");
    }

    unguard();
}

// L2CacheD 0x00482740
bool CIOServer::Create(int port)
{
    guard(L"BOOL CIOServer::Create(int nPort)", L"BOOL CIOServer::Create(int nPort) (exit)", TEXT(__FILE__), __LINE__);

    m_hSocket = ::socket(AF_INET, AF_UNIX, IPPROTO_IP);
    if (m_hSocket == INVALID_SOCKET)
    {
        int err = ::WSAGetLastError();
        g_winlog.Add_Ansi(LOG_ERROR, "socket error %d", err);
        unguard();
        return false;
    }

    sockaddr_in name;
    name.sin_family = AF_INET;
    name.sin_addr.s_addr = ::htonl(INADDR_ANY);
    name.sin_port = ::htons(port);

    if (::bind(m_hSocket, (const struct sockaddr*)&name, sizeof(sockaddr_in)))
    {
        int err = ::WSAGetLastError();
        g_winlog.Add_Ansi(LOG_ERROR, "bind error %d", err);

        Close();
        unguard();
        return false;
    }

    if (::listen(m_hSocket, 5))
    {
        int err = ::WSAGetLastError();
        g_winlog.Add_Ansi(LOG_ERROR, "listen error %d", err);

        Close();
        unguard();
        return false;
    }

    m_hAcceptEvent = ::WSACreateEvent();
    ::WSAEventSelect(m_hSocket, m_hAcceptEvent, FD_ACCEPT);

    if (!RegisterWait(m_hAcceptEvent))
    {
        g_winlog.Add_Ansi(LOG_ERROR, "RegisterWait error on port %d", port);

        Close();
        unguard();
        return false;
    }

    unguard();

    return true;
}

SOCKET CIOServer::getSocket() const
{
    return m_hSocket;
}

// L2CacheD 0x0047F870
void CIOServer::Close()
{
    guard(L"void CIOServer::Close()", L"void CIOServer::Close() (exit)", TEXT(__FILE__), __LINE__);

    if (m_hSocket != INVALID_SOCKET)
    {
        ::closesocket(m_hSocket);
        m_hSocket = INVALID_SOCKET;
    }

    if (m_hAcceptEvent != WSA_INVALID_EVENT)
    {
        ::WSACloseEvent(m_hAcceptEvent);
        m_hAcceptEvent = WSA_INVALID_EVENT;
    }

    unguard();
}
